{# -*- engine: jinja -*- #}
<!DOCTYPE html>
<html class="theme-{{theme}}">

  <head>
    <meta http-equiv="Content-Type" content="text/html; charset={{info.encoding}}"/>
    <title>{{info.head}}</title>
    <meta name="viewport" content="width=device-width, initial-scale=1.0"/>
    {% if css_link is defined %}
    <link rel="stylesheet" href="{{css_link}}"/>
    {% else %}
    <style type="text/css">
{{ css | safe }}
    </style>
    {% endif %}
  </head>

  <body{% if info.single_page == "js-enabled" %} class="js-disabled"{% endif %}>

    <header>
      <h1>{{info.head}}</h1>
      <hr/>
      {% if not info.single_page %}
      <div class="summary">
      {% endif %}
        {% block summary %}
        {% endblock %}
      {% if not info.single_page %}
      </div>
      {% endif %}
      <hr/>
    </header>
    <nav>
      {% block navigation %}
      {% endblock %}
    </nav>
    <main>
      {% block content %}
      {% endblock %}
      <hr/>
    </main>
    <footer>
      Generated by: <a href="http://gcovr.com/en/{{ 'main' if info.version.endswith('+main') else info.version }}">GCOVR (Version {{ info.version }})</a>
    </footer>
  </body>
</html>
