{# -*- engine: jinja -*- #}
{% if function_list %}
<details>
  <summary>List of functions</summary>
  <table class="listOfFunctions">
    <tr>
      <th>Function (Line)</th>
      <th>Call count</th>
      {% if not EXCLUDE_FUNCTION_COVERAGE %}
      <th>Line coverage</th>
      <th>Branch coverage</th>
      <th>Condition coverage</th>
      {% endif %}
      <th>Block coverage</th>
    </tr>
{% if info.single_page %}
{%  set anchor_prefix = html_filename + "|" %}
{% else %}
{%  set anchor_prefix = '' %}
{% endif %}
{% for entry in function_list %}
    <tr>
      {#- #}<td><a href="#{{ anchor_prefix }}l{{ entry['line'] }}">{{ entry["name"] }} (line {{ entry["line"] }})</a></td>
      {#- #}<td>
                {%- if entry["excluded"] %}excluded
                {%- else %}
                {%-  if entry["count"] == 0 %}not called
                {%-  else %}called {{ entry["count"] }} time{% if entry["count"] > 1 %}s{% endif %}
                {%-  endif -%}
                {%- endif -%}
      {#- #}</td>
      {%- if not EXCLUDE_FUNCTION_COVERAGE %}
      {#- #}<td>{{ entry["line_coverage"] }}%</td>
      {#- #}<td>{{ entry["branch_coverage"] }}%</td>
      {#- #}<td>{{ entry["condition_coverage"] }}%</td>
      {%- endif %}
      {#- #}<td>{{ entry["blocks"] }}%</td>
      {#- -#}
    </tr>
{% endfor %}
  </table>
</details>
{% endif -%}
