{# -*- engine: jinja -*- #}
<!DOCTYPE html>
<html class="theme-{{theme}}" {% if (theme == 'dark-blue' or theme == 'dark-green') %}data-color-mode="dark" data-dark-theme="dark"{% endif %}>
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset={{info.encoding}}"/>
    <title>{{info.head}}</title>
    <meta name="viewport" content="width=device-width, initial-scale=1.0"/>
    {% if css_link is defined %}
    <link rel="stylesheet" href="{{css_link}}"/>
    {% else %}
    <style type="text/css">
{{ css | safe }}
    </style>
    {% endif %}
  </head>
  <body{% if info.single_page == "js-enabled" %} class="js-disabled"{% endif %}>
    <div class="Layout-main">
      <div class="Layout-main-centered-lg" style="max-width: 1280px;">
        <div class="d-flex flex-justify-center">
          <h1>{{info.head}}</h1>
        </div>
        <hr/>
        {% if not info.single_page %}
        <div class="summary">
        {% endif %}
          {% block summary %}
          {% endblock %}
        {% if not info.single_page %}
        </div>
        {% endif %}
        <hr/>
        <nav>
          {% block navigation %}
          {% endblock %}
        </nav>
        <main>
          {% block content %}
          {% endblock %}
          <hr/>
        </main>
        <footer class="text-center">
          Generated by: <a href="http://gcovr.com/en/{{ 'main' if info.version.endswith('+main') else info.version }}">GCOVR (Version {{ info.version }})</a>
        </footer>
      </div>
    </div>
  </body>
</html>
