:root {
    font-family: sans-serif;
    --tab_size: {{tab_size}};
}

.theme-green, .theme-blue {
    --unknown_color: lightgray;
    --low_color: #FF6666;
    --medium_color: #F9FD63;
    --partial_covered_color: var(--medium_color);
    --uncovered_color: #FF8C8C;
    --warning_color: orangered;
    --notTakenBranch_color: red;
    --notTakenCondition_color: red;
    --uncheckedDecision_color: darkorange;
    --notTakenDecision_color: red;
    --notInvokedCall_color: red;
    --excluded_color: rgb(255, 241, 229);
}

.theme-green {
    --high_color: #85E485;
    --covered_color: #85E485;
    --takenBranch_color: green;
    --takenCondition_color: green;
    --takenDecision_color: green;
    --invokedCall_color: green;
}

.theme-blue {
    --high_color: #66B4FF;
    --covered_color: #66B4Ff;
    --takenBranch_color: blue;
    --takenCondition_color: blue;
    --takenDecision_color: blue;
    --invokedCall_color: blue;
}

body
{
    color: #000000;
    background-color: #FFFFFF;
}

h1
{
    text-align: center;
    margin: 0;
    padding-bottom: 10px;
    font-size: 20pt;
    font-weight: bold;
}

hr
{
    background-color: navy;
    height: 2px;
    border: 0;
}

/* Link formats: use maroon w/underlines */
a:link
{
    color: navy;
    text-decoration: underline;
}
a:visited
{
    color: maroon;
    text-decoration: underline;
}

{% include "style.summary.css" %}

{% include "style.meter.css" %}

{% include "style.filelist.css" %}

{% include "style.details.css" %}

{% if single_page == "js-enabled" %}
.js-enabled .js-enabled-hidden {
    display: none
}

.js-disabled .js-disabled-hidden {
    display: none
}

{% endif %}
footer
{
    text-align: center;
    padding-top: 3px;
}
